#!/usr/bin/python3
import os
import sys
import logging
import socket
import socketserver

import click
import requests

logger = logging.getLogger(__name__)

class UffdAPI:
	def __init__(self, baseurl, client_id, client_secret):
		self.baseurl = baseurl
		self.session = requests.Session()
		self.session.auth = (client_id, client_secret)

	def get(self, endpoint, **kwargs):
		resp = self.session.get(self.baseurl + endpoint, params=kwargs)
		resp.raise_for_status()
		return resp.json()

	def get_aliases(self, mail_address):
		'''Return list of alias addresses for a given mail address'''
		results = self.get('/api/v1/getmails', receive_address=mail_address)
		destinations = []
		for result in results:
			destinations += result['destination_addresses']
		logger.debug('API getmails response for %s: %s', repr(mail_address), destinations)
		return destinations

	def resolve_remailer(self, orig_address):
		'''Return real mail address for a given remailer mail address or None'''
		result = self.get('/api/v1/resolve-remailer', orig_address=orig_address)
		return result['address']

# From https://cr.yp.to/proto/netstrings.txt
#
# Any string of 8-bit bytes may be encoded as [len]":"[string]",".
# Here [string] is the string and [len] is a nonempty sequence of ASCII
# digits giving the length of [string] in decimal. The ASCII digits are
# <30> for 0, <31> for 1, and so on up through <39> for 9. Extra zeros
# at the front of [len] are prohibited: [len] begins with <30> exactly
# when [string] is empty.
#
# For example, the string "hello world!" is encoded as <31 32 3a 68
# 65 6c 6c 6f 20 77 6f 72 6c 64 21 2c>, i.e., "12:hello world!,". The
# empty string is encoded as "0:,".
#
# [len]":"[string]"," is called a netstring. [string] is called the
# interpretation of the netstring.

def encode_netstring(string):
	'''Return netstring encoding of a Python str as bytes'''
	string = string.encode('utf8')
	return b'%d:%s,'%(len(string), string)

def _parse_netstring_length(netstring):
	if b':' not in netstring:
		raise ValueError()
	length, rest = netstring.split(b':', 1)
	if not length.isdigit():
		raise ValueError()
	length = int(length)
	return length, rest

def is_netstring_complete(netstring):
	'''Return whether the argument (bytes) is a complete netstring

	Raises ValueError if the argument is invalid instead of only incomplete.'''
	if b':' not in netstring:
		return False
	length, rest = _parse_netstring_length(netstring)
	return len(rest) >= length + 1

def decode_netstring(netstring):
	'''Decode netstring at the beginning of argument

	Returns (decoded_str, remaining_bytes). Raises ValueError if argument does
	not start with a valid and complete netstring or the encoded string is not
	valid UTF-8.'''
	length, rest = _parse_netstring_length(netstring)
	string, comma, rest = rest[:length], rest[length:length+1], rest[length+1:]
	if len(string) != length or comma != b',':
		raise ValueError()
	# UnicodeDecodeError is a subclass of ValueError
	return string.decode('utf8'), rest

# From socketmap_table(5)
#
# PROTOCOL
# Socketmaps use a simple protocol: the client sends one request, and the
# server sends one reply. Each request and each reply are sent as one
# netstring object.
#
# REQUEST FORMAT
# The socketmap protocol supports only the lookup request. The request
# has the following form:
#
# name <space> key
#     Search the named socketmap for the specified key.
#
# [...]
# REPLY FORMAT
# The Postfix socketmap client requires that replies are not longer than
# 100000 characters (not including the netstring encapsulation). Replies
# must have the following form:
#
# OK <space> data
#     The requested data was found.
#
# NOTFOUND <space>
#     The requested data was not found.
#
# TEMP <space> reason
#
# TIMEOUT <space> reason
#
# PERM <space> reason
#     The request failed. The reason, if non-empty, is descriptive
#     text.

class SocketmapsRequestHandler(socketserver.BaseRequestHandler):
	api = None # Overwritten

	def handle(self):
		buf = b''
		while not is_netstring_complete(buf):
			chunk = self.request.recv(4096)
			if not chunk:
				return
			buf += chunk
		request, rest = decode_netstring(buf)
		if rest:
			return
		try:
			response = self.handle_request(request)
		except: # pylint: disable=bare-except
			logger.exception('Exception during request')
			self.request.sendall(encode_netstring('TEMP Internal Error'))
			return
		loglevel = logging.INFO
		if isinstance(response, tuple) and len(response) == 2:
			response, loglevel = response
		logger.log(loglevel, 'Request %s -> %s', repr(request), repr(response))
		self.request.sendall(encode_netstring(response))

	def handle_request(self, request):
		if ' ' not in request:
			return 'PERM Malformed request', logging.WARNING
		name, key = request.split(' ', 1)
		if name == 'virtual':
			# Postfix option virtual_alias_domains defaults to virtual_alias_maps,
			# so we might get some domain requests. Since Postfix does not retry
			# failing virtual_alias_domains lookups, PERM is the best option here.
			if '@' not in key:
				return 'PERM Key is not a valid mail address. Maybe fix the virtual_alias_domains setting.', logging.WARNING
			key = key.strip().lower()
			# Emulate Postfixes virtual(8) behaviour when used with ldap_table(5).
			local, domain = key.split('@', 1)
			if '+' in local:
				local, extention = local.split('+', 1)
				lookup_order = [
					(f'{local}+{extention}@{domain}', ''),
					(f'{local}@{domain}', extention),
					(f'{local}+{extention}', ''),
					(f'{local}', extention),
					(f'@{domain}', extention),
				]
			else:
				lookup_order = [
					(f'{local}@{domain}', ''),
					(f'{local}', ''),
					(f'@{domain}', ''),
				]
			results = []
			for key, extention in lookup_order:
				for address in self.api.get_aliases(key):
					# All destinations must be full addresses, but we don't want to fail
					# here, if any is malformed.
					if '@' in address and extention:
						local, domain = address.split('@', 1)
						address = f'{local}+{extention}@{domain}'
					results.append(address)
				if results:
					break
			# Returning "OK " (without any data) is interperted as a protocol error
			# and Postfix will retry the request.
			if not results:
				return 'NOTFOUND '
			# There is no way to escape commas in the result items and it would not
			# matter, since Postfix internally processes lookup results as a
			# comma-separated list.
			return 'OK ' + ','.join(results)
		if name == 'remailer_canonical':
			domain = key.rsplit('@', 1)[-1]
			result = self.api.resolve_remailer(key)
			if result is None:
				return 'NOTFOUND '
			return 'OK ' + result
		return 'PERM Unknown request name', logging.WARNING

def make_requesthandler(api):
	class RequestHandler(SocketmapsRequestHandler):
		pass
	RequestHandler.api = api
	return RequestHandler

class FilenoUnixStreamServer(socketserver.UnixStreamServer):
	def __init__(self, fd, RequestHandlerClass, bind_and_activate=True):
		self.server_fd = fd
		super().__init__(None, RequestHandlerClass, bind_and_activate=bind_and_activate)

	def server_bind(self):
		self.socket.close() # UnixStreamServer.__init__ creates an unbound socket
		self.socket = socket.fromfd(self.server_fd, socket.AF_UNIX, socket.SOCK_STREAM)
		self.server_address = self.socket.getsockname()

class ThreadingFilenoUnixStreamServer(socketserver.ThreadingMixIn, FilenoUnixStreamServer):
	pass

def cleanup_unix_socket(path):
	if not os.path.exists(path):
		return
	conn = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	try:
		conn.connect(path)
	except ConnectionRefusedError:
		os.remove(path)
	conn.close()

class StdoutFilter(logging.Filter):
	def filter(self, record):
		return record.levelno <= logging.INFO

@click.command(help='Socketmap proxy for integrating Postfix and other compatible MTAs with uffd SSO. Supports virtual alias lookups (request name "virtual").')
@click.option('--socket-path', type=click.Path(), help='Path for UNIX domain socket')
@click.option('--socket-fd', type=int, help='Use fd number as server socket (alternative to --socket-path)')
@click.option('--api-url', required=True, help='Uffd base URL without API prefix or trailing slash (e.g. https://example.com)')
@click.option('--api-user', required=True, help='API user/client id')
@click.option('--api-secret', required=True, help='API secret, do not set this on the command-line, use environment variable SERVER_API_SECRET instead')
def main(socket_path, socket_fd, api_url, api_user, api_secret):
	if (socket_path is None and socket_fd is None) or \
	   (socket_path is not None and socket_fd is not None):
		raise click.ClickException('Either --socket-path or --socket-fd must be specified')

	stdout_handler = logging.StreamHandler(sys.stdout)
	stdout_handler.setLevel(logging.DEBUG)
	stdout_handler.addFilter(StdoutFilter())
	stderr_handler = logging.StreamHandler(sys.stderr)
	stderr_handler.setLevel(logging.WARNING)
	logger.setLevel(logging.INFO)
	logger.addHandler(stdout_handler)
	logger.addHandler(stderr_handler)

	api = UffdAPI(api_url, api_user, api_secret)
	RequestHandler = make_requesthandler(api)
	if socket_path is not None:
		cleanup_unix_socket(socket_path)
		socketserver.ThreadingUnixStreamServer(socket_path, RequestHandler).serve_forever()
	else:
		ThreadingFilenoUnixStreamServer(socket_fd, RequestHandler).serve_forever()

if __name__ == '__main__':
	# Pylint does not seem to understand the click's decorators
	# pylint: disable=unexpected-keyword-arg,no-value-for-parameter
	main(auto_envvar_prefix='SERVER')
